# Copyright (c) 2025 Patricio Cubillos
# Gen TSO is open-source software under the GPL-2.0 license (see LICENSE)

import os
import pickle
import pytest

import numpy as np

import gen_tso.pandeia_io as jwst
from gen_tso.utils import ROOT

os.chdir(ROOT+'../tests')
# See tests/mocks/make_mocks.py for mocked data setup.


def test_saturation_level_perform_calculation_single():
    with open('mocks/perform_calculation_nircam_lw_tsgrism.pkl', 'rb') as f:
        result = pickle.load(f)

    pixel_rate, full_well = jwst.saturation_level(result)
    expected_rate = 1300.2144775390625
    expected_well = 58100.00422843957
    np.testing.assert_allclose(pixel_rate, expected_rate)
    np.testing.assert_allclose(full_well, expected_well)


def test_saturation_level_perform_calculation_multi():
    with open('mocks/perform_calculation_miri_mrs_ts.pkl', 'rb') as f:
        result = pickle.load(f)

    pixel_rate, full_well = jwst.saturation_level(result)
    expected_rate = [
        204.37779236, 109.17165375,  32.75593948,   4.55676889,
    ]
    expected_well = [
        193654.99564498, 193655.00540975, 193655.004117  , 193654.99798687,
    ]
    np.testing.assert_allclose(pixel_rate, expected_rate)
    np.testing.assert_allclose(full_well, expected_well)


def test_saturation_level_tso_calculation_single():
    with open('mocks/tso_calculation_nircam_lw_tsgrism.pkl', 'rb') as f:
        tso = pickle.load(f)

    pixel_rate, full_well = jwst.saturation_level(tso)
    expected_rate = [1261.23925781, 1300.21447754]
    expected_well = [58100.00016363, 58100.00129109]
    np.testing.assert_allclose(pixel_rate, expected_rate)
    np.testing.assert_allclose(full_well, expected_well)


def test_saturation_level_tso_calculation_multi():
    with open('mocks/tso_calculation_miri_mrs_ts.pkl', 'rb') as f:
        tso = pickle.load(f)

    pixel_rate, full_well = jwst.saturation_level(tso)
    expected_rate = [
        198.24740601, 105.90003204,  31.80119514,   4.46783495,
        204.37779236, 109.17165375,  32.75593948,   4.55676889,
    ]
    expected_well = [
        193654.99147199, 193654.99886718, 193654.98493823, 193654.99820043,
        193654.99564498, 193655.00540975, 193655.004117  , 193654.99798687,
    ]
    np.testing.assert_allclose(pixel_rate, expected_rate)
    np.testing.assert_allclose(full_well, expected_well)


@pytest.mark.skip(reason='TBD')
def test_saturation_level_tso_calculation_get_max():
    pass


@pytest.mark.parametrize('nint', [1, 10, 100])
def test_exposure_time_nircam(nint):
    inst = 'nircam'
    subarray = 'subgrism64'
    readout = 'rapid'
    ngroup = 90
    exp_time = jwst.exposure_time(inst, subarray, readout, ngroup, nint)
    if nint == 1:
        expected_exposure = 31.00063
    elif nint == 10:
        expected_exposure = 310.0063
    elif nint == 100:
        expected_exposure = 3100.063
    np.testing.assert_allclose(exp_time, expected_exposure)


@pytest.mark.parametrize('nint', [1, 10, 100])
def test_exposure_time_miri_fastr1(nint):
    inst = 'miri'
    subarray = 'slitlessprism'
    readout = 'fastr1'
    ngroup = 30
    exp_time = jwst.exposure_time(inst, subarray, readout, ngroup, nint)
    if nint == 1:
        expected_exposure = 4.7712
    elif nint == 10:
        expected_exposure = 49.14336
    elif nint == 100:
        expected_exposure = 492.86496
    np.testing.assert_allclose(exp_time, expected_exposure)


@pytest.mark.parametrize('nint', [1, 10, 100])
def test_exposure_time_miri_slowr1(nint):
    inst = 'miri'
    subarray = 'full'
    readout = 'slowr1'
    ngroup = 30
    exp_time = jwst.exposure_time(inst, subarray, readout, ngroup, nint)
    if nint == 1:
        expected_exposure = 716.6976
    elif nint == 10:
        expected_exposure = 7381.98528
    elif nint == 100:
        expected_exposure = 74034.86208
    np.testing.assert_allclose(exp_time, expected_exposure)


def test_exposure_time_bad_subarray():
    inst = 'nircam'
    subarray = 'nope'
    readout = 'rapid'
    ngroup = 90
    nint = 1
    exp_time = jwst.exposure_time(inst, subarray, readout, ngroup, nint)
    assert exp_time == 0.0


def test_exposure_time_bad_readout():
    inst = 'nircam'
    subarray = 'subgrism64'
    readout = 'nope'
    ngroup = 90
    nint = 1
    exp_time = jwst.exposure_time(inst, subarray, readout, ngroup, nint)
    assert exp_time == 0.0


def test_bin_search_exposure_time_nircam():
    inst = 'nircam'
    subarray = 'subgrism64'
    readout = 'rapid'
    ngroup = 90
    obs_time = 6.0
    nint, exp_time = jwst.bin_search_exposure_time(
        inst, subarray, readout, ngroup, obs_time,
    )
    assert nint == 697
    np.testing.assert_allclose(exp_time, 21607.43911)


def test_bin_search_exposure_time_miri():
    inst = 'miri'
    subarray = 'slitlessprism'
    readout = 'fastr1'
    ngroup = 30
    obs_time = 6.0
    nint, exp_time = jwst.bin_search_exposure_time(
        inst, subarray, readout, ngroup, obs_time,
    )
    assert nint == 4382
    np.testing.assert_allclose(exp_time, 21604.15264)


def test_bin_search_exposure_time_bad_subarray():
    inst = 'nircam'
    subarray = 'nope'
    readout = 'rapid'
    ngroup = 90
    obs_time = 6.0
    nint, exp_time = jwst.bin_search_exposure_time(
        inst, subarray, readout, ngroup, obs_time,
    )
    assert exp_time == 0.0


def test_bin_search_exposure_time_bad_readout():
    inst = 'nircam'
    subarray = 'subgrism64'
    readout = 'nope'
    ngroup = 90
    obs_time = 6.0
    nint, exp_time = jwst.bin_search_exposure_time(
        inst, subarray, readout, ngroup, obs_time,
    )
    assert exp_time == 0.0


def test__print_pandeia_exposure_config():
    with open('mocks/perform_calculation_nircam_lw_tsgrism.pkl', 'rb') as f:
        result = pickle.load(f)
    config = result['input']['configuration']
    text = jwst._print_pandeia_exposure(config=config)
    assert text == 'Exposure time: 701.41 s (0.19 h)'


def test__print_pandeia_exposure_vals():
    inst = 'nircam'
    subarray = 'subgrism64'
    readout = 'rapid'
    ngroup = 90
    nint = 150
    text = jwst._print_pandeia_exposure(inst, subarray, readout, ngroup, nint)
    assert text == 'Exposure time: 4650.09 s (1.29 h)'


@pytest.mark.parametrize('format', [None, 'html', 'rich'])
def test__print_pandeia_exposure_format(format):
    with open('mocks/perform_calculation_nircam_lw_tsgrism.pkl', 'rb') as f:
        result = pickle.load(f)
    config = result['input']['configuration']
    text = jwst._print_pandeia_exposure(config=config, format=format)
    assert text == 'Exposure time: 701.41 s (0.19 h)'


def test__print_pandeia_saturation_perform_calc():
    with open('mocks/perform_calculation_nircam_lw_tsgrism.pkl', 'rb') as f:
        result = pickle.load(f)

    text = jwst._print_pandeia_saturation(reports=[result], format=None)
    expected_text = """Max fraction of saturation: 1.5%
ngroup below 80% saturation: 104
ngroup below 100% saturation: 131"""
    assert text == expected_text


def test__print_pandeia_saturation_tso_calc():
    with open('mocks/tso_calculation_miri_mrs_ts.pkl', 'rb') as f:
        tso = pickle.load(f)

    text = jwst._print_pandeia_saturation(reports=tso, format=None)
    expected_text = """Max fraction of saturation: 73.2%
ngroup below 80% saturation: 273
ngroup below 100% saturation: 341"""
    assert text == expected_text


def test__print_pandeia_saturation_values():
    pixel_rate, full_well = 1396.5528564453125, 58100.002280515175
    inst = 'nircam'
    subarray = 'subgrism64'
    readout = 'rapid'
    ngroup = 90
    text = jwst._print_pandeia_saturation(
        inst, subarray, readout, ngroup, pixel_rate, full_well,
        format=None,
    )
    expected_text = """Max fraction of saturation: 73.7%
ngroup below 80% saturation: 97
ngroup below 100% saturation: 122"""
    assert text == expected_text


def test__print_pandeia_saturation_req_saturation():
    pixel_rate, full_well = 1396.5528564453125, 58100.002280515175
    inst = 'nircam'
    subarray = 'subgrism64'
    readout = 'rapid'
    ngroup = 90
    text = jwst._print_pandeia_saturation(
        inst, subarray, readout, ngroup, pixel_rate, full_well,
        format=None,
        req_saturation=70.0,
    )
    expected_text = """Max fraction of saturation: 73.7%
ngroup below 70% saturation: 85
ngroup below 100% saturation: 122"""
    assert text == expected_text


@pytest.mark.skip(reason='TBD')
def test__print_pandeia_stats():
    pass


def test__print_pandeia_report_perform_calculation_single():
    with open('mocks/perform_calculation_nircam_lw_tsgrism.pkl', 'rb') as f:
        result = pickle.load(f)

    report = jwst._print_pandeia_report([result], format=None)
    expected_report = """Exposure time: 701.41 s (0.19 h)
Max fraction of saturation: 1.5%
ngroup below 80% saturation: 104
ngroup below 100% saturation: 131

Signal-to-noise ratio        363.4
Extracted flux              1815.8  e-/s
Flux standard deviation        5.0  e-/s
Brightest pixel rate        1300.2  e-/s

Integrations:                         683
Duty cycle:                          0.66
Total exposure time:                701.4  s
First--last dt per exposure:        701.4  s
Reset--last dt per integration:     232.6  s

Reference wavelength:                    4.46  microns
Area of extraction aperture:             4.76  pixels
Area of background measurement:           6.3  pixels
Background surface brightness:            0.3  MJy/sr
Total sky flux in background aperture:   4.94  e-/s
Total flux in background aperture:      59.59  e-/s
Background flux fraction from scene:     0.92
Number of cosmic rays:      0.0002  events/pixel/read"""
    assert report == expected_report



def test__print_pandeia_report_perform_calculation_multi():
    with open('mocks/perform_calculation_miri_mrs_ts.pkl', 'rb') as f:
        result = pickle.load(f)

    report = jwst._print_pandeia_report(result, format=None)
    expected_report = """Exposure time: 27858.63 s (7.74 h)
Max fraction of saturation: 73.2%
ngroup below 80% saturation: 273
ngroup below 100% saturation: 341

Signal-to-noise ratio       1334.2    1265.2     960.1     179.2
Extracted flux               225.1     147.4      46.2       2.0  e-/s
Flux standard deviation        0.2       0.1       0.0       0.0  e-/s
Brightest pixel rate         204.4     109.2      32.8       4.6  e-/s

Integrations:                          40
Duty cycle:                          0.99
Total exposure time:              27858.6  s
First--last dt per exposure:      27858.6  s
Reset--last dt per integration:   27639.4  s

Reference wavelength:                    5.35   8.15  12.50  19.29  microns
Area of extraction aperture:             8.20   5.21   2.98   1.61  pixels
Area of background measurement:          79.4   50.5   28.9   15.6  pixels
Background surface brightness:            1.0   10.9   43.0  165.5  MJy/sr
Total sky flux in background aperture:   0.29   0.82   2.98   2.44  e-/s
Total flux in background aperture:       0.85   1.47   3.19   2.54  e-/s
Background flux fraction from scene:     0.66   0.45   0.06   0.04
Number of cosmic rays:      0.3122  events/pixel/read"""
    assert report == expected_report


def test__print_pandeia_report_tso_calculation_single():
    with open('mocks/tso_calculation_nircam_lw_tsgrism.pkl', 'rb') as f:
        result = pickle.load(f)

    report = jwst._print_pandeia_report(result, format=None)
    expected_report = """Exposure time: 21607.44 s (6.00 h)
Max fraction of saturation: 68.6%
ngroup below 80% saturation: 104
ngroup below 100% saturation: 131

Signal-to-noise ratio       3241.5
Extracted flux              1761.3  e-/s
Flux standard deviation        0.5  e-/s
Brightest pixel rate        1261.2  e-/s

                               in-transit  out-transit
Integrations:                         244      453
Duty cycle:                          0.99     0.99
Total exposure time:               7564.2  14043.3  s
First--last dt per exposure:       7564.2  14043.3  s
Reset--last dt per integration:    7396.7  13732.4  s

Reference wavelength:                    4.46  microns
Area of extraction aperture:             4.76  pixels
Area of background measurement:           6.3  pixels
Background surface brightness:            0.3  MJy/sr
Total sky flux in background aperture:   4.94  e-/s
Total flux in background aperture:      57.95  e-/s
Background flux fraction from scene:     0.91
Number of cosmic rays:      0.0072  events/pixel/read"""
    assert report == expected_report


def test__print_pandeia_report_tso_calculation_multi():
    with open('mocks/tso_calculation_miri_mrs_ts.pkl', 'rb') as f:
        tso = pickle.load(f)

    report = jwst._print_pandeia_report(tso, format=None)
    expected_report = """Exposure time: 20893.28 s (5.80 h)
Max fraction of saturation: 73.2%
ngroup below 80% saturation: 273
ngroup below 100% saturation: 341

Signal-to-noise ratio        657.0     623.0     472.3      87.5
Extracted flux               218.4     142.9      44.8       1.9  e-/s
Flux standard deviation        0.3       0.2       0.1       0.0  e-/s
Brightest pixel rate         198.2     105.9      31.8       4.5  e-/s

                               in-transit  out-transit
Integrations:                          10       20
Duty cycle:                          0.99     0.99
Total exposure time:               6962.6  13927.9  s
First--last dt per exposure:       6962.6  13927.9  s
Reset--last dt per integration:    6909.8  13819.7  s

Reference wavelength:                    5.35   8.15  12.50  19.29  microns
Area of extraction aperture:             8.20   5.21   2.98   1.61  pixels
Area of background measurement:          79.4   50.5   28.9   15.6  pixels
Background surface brightness:            1.0   10.9   43.0  165.5  MJy/sr
Total sky flux in background aperture:   0.29   0.82   2.98   2.44  e-/s
Total flux in background aperture:       0.83   1.45   3.18   2.54  e-/s
Background flux fraction from scene:     0.65   0.44   0.06   0.04
Number of cosmic rays:      0.3122  events/pixel/read"""
    assert report == expected_report


def test_tso_print_plain(capsys):
    with open('mocks/tso_calculation_nircam_lw_tsgrism.pkl', 'rb') as f:
        result = pickle.load(f)
    # Make some troublesome figures
    result['report_in']['scalar']['duty_cycle'] = 0.25
    result['report_in']['input']['configuration']['detector']['ngroup'] = 130
    result['report_out']['scalar']['duty_cycle'] = 0.25
    result['report_out']['input']['configuration']['detector']['ngroup'] = 130

    jwst.tso_print(result, format=None)
    captured = capsys.readouterr()
    expected_report = (
        "Exposure time: 31103.65 s (8.64 h)\r\n"
        "Max fraction of saturation: 99.1%\r\n"
        "ngroup below 80% saturation: 104\r\n"
        "ngroup below 100% saturation: 131\r\n"
        "\r\n"
        "Signal-to-noise ratio       3241.5\r\n"
        "Extracted flux              1761.3  e-/s\r\n"
        "Flux standard deviation        0.5  e-/s\r\n"
        "Brightest pixel rate        1261.2  e-/s\r\n"
        "\r\n"
        "                               in-transit  out-transit\r\n"
        "Integrations:                         244      453\r\n"
        "Duty cycle:                          0.25     0.25\r\n"
        "Total exposure time:               7564.2  14043.3  s\r\n"
        "First--last dt per exposure:       7564.2  14043.3  s\r\n"
        "Reset--last dt per integration:    7396.7  13732.4  s\r\n"
        "\r\n"
        "Reference wavelength:                    4.46  microns\r\n"
        "Area of extraction aperture:             4.76  pixels\r\n"
        "Area of background measurement:           6.3  pixels\r\n"
        "Background surface brightness:            0.3  MJy/sr\r\n"
        "Total sky flux in background aperture:   4.94  e-/s\r\n"
        "Total flux in background aperture:      57.95  e-/s\r\n"
        "Background flux fraction from scene:     0.91\r\n"
        "Number of cosmic rays:      0.0072  events/pixel/read\r\n"
    )
    assert captured.out == expected_report


@pytest.mark.skip(reason='Having some issues with capsys IO')
def test_tso_print_rich(capsys):
    with open('mocks/tso_calculation_nircam_lw_tsgrism.pkl', 'rb') as f:
        result = pickle.load(f)
    # Make some troublesome figures
    result['report_in']['scalar']['duty_cycle'] = 0.25
    result['report_in']['input']['configuration']['detector']['ngroup'] = 130
    result['report_out']['scalar']['duty_cycle'] = 0.25
    result['report_out']['input']['configuration']['detector']['ngroup'] = 130

    jwst.tso_print(result)
    captured = capsys.readouterr()
    expected_report = (
        "Exposure time: 31014.40 s (8.62 h)\r\n"
        "Max fraction of saturation: 106.4%\r\n"
        "ngroup below 80% saturation: 97\r\n"
        "ngroup below 100% saturation: 122\r\n"
        "\r\n"
        "Signal-to-noise ratio       3484.8\r\n"
        "Extracted flux              2043.0  e-/s\r\n"
        "Flux standard deviation        0.6  e-/s\r\n"
        "Brightest pixel rate        1354.7  e-/s\r\n"
        "\r\n"
        "                               in-transit  out-transit\r\n"
        "Integrations:                         243      452\r\n"
        "Duty cycle:                          0.25     0.25\r\n"
        "Total exposure time:               7533.2  14012.3  s\r\n"
        "First--last dt per exposure:       7533.2  14012.3  s\r\n"
        "Reset--last dt per integration:    7366.4  13702.1  s\r\n"
        "\r\n"
        "Reference wavelength:                    4.36  microns\r\n"
        "Area of extraction aperture:             4.76  pixels\r\n"
        "Area of background measurement:           6.3  pixels\r\n"
        "Background surface brightness:            0.3  MJy/sr\r\n"
        "Total sky flux in background aperture:   4.45  e-/s\r\n"
        "Total flux in background aperture:      64.12  e-/s\r\n"
        "Background flux fraction from scene:     0.93\r\n"
        "Number of cosmic rays:      0.0072  events/pixel/read\r\n"
    )
    assert captured.out == expected_report


def test_simulate_tso_spectroscopy():
    if False:
        import gen_tso.pandeia_io as jwst
        from gen_tso.utils import ROOT

        pando = jwst.PandeiaCalculation('nirspec', 'bots')
        pando.set_scene('phoenix', 'k5v', '2mass,ks', 8.351)

        disperser='g395h'
        filter='f290lp'
        readout='nrsrapid'
        subarray='sub2048'
        ngroup = 14

        transit_dur = 2.71
        obs_dur = 6.01
        obs_type = 'transit'
        model_path = f'{ROOT}data/models//WASP80b_transit.dat'
        depth_model = np.loadtxt(model_path, unpack=True)

        tso = pando.tso_calculation(
            obs_type, transit_dur, obs_dur, depth_model,
            ngroup, disperser, filter, subarray, readout,
        )
        pando.save_tso(filename='mocks/tso_nirspec.pickle')

    bin_wl, bin_spec, bin_err, bin_widths = jwst.simulate_tso(
       tso, n_obs=1, resolution=50.0, noiseless=False,
    )

    expected_bin_wl = np.array([
       2.89953946, 2.95811601, 3.01787593, 3.07884313, 3.14104198,
       3.20449737, 3.26923469, 3.33527984, 3.40265923, 3.47139982,
       3.54152911, 3.61307515, 3.68606657, 3.83650291, 3.91400802,
       3.99307889, 4.07374715, 4.15604507, 4.24000558, 4.32566226,
       4.41304937, 4.50220189, 4.59315546, 4.68594648, 4.78061206,
       4.87719009, 4.97571918, 5.07623876, 5.15183947])

    expected_bin_spec = np.array([
       0.0297344 , 0.02974732, 0.02965458, 0.02969755, 0.02970218,
       0.02979896, 0.02968281, 0.02986269, 0.02970917, 0.02964738,
       0.02957673, 0.02946511, 0.02929959, 0.02917717, 0.02907652,
       0.02908018, 0.02911588, 0.02916655, 0.02934231, 0.02940869,
       0.02940897, 0.02928256, 0.02923854, 0.02929097, 0.02931623,
       0.02939239, 0.02943399, 0.02943881, 0.02939068])

    expected_bin_err = np.array([
       2.08823769e-05, 1.94691040e-05, 1.85650131e-05, 1.80149594e-05,
       1.78999841e-05, 1.79079847e-05, 1.76830162e-05, 1.77418014e-05,
       1.78654463e-05, 1.78222581e-05, 1.80192033e-05, 1.83102194e-05,
       1.91776674e-05, 2.41068417e-05, 2.02616020e-05, 2.10061215e-05,
       2.17422060e-05, 2.24678539e-05, 2.33846872e-05, 2.53903572e-05,
       2.74600911e-05, 2.89515184e-05, 3.04996226e-05, 3.15485271e-05,
       3.30016963e-05, 3.42489607e-05, 3.59747181e-05, 3.71223378e-05,
       5.52240362e-05])

    expected_bin_widths = np.array([
       0.02899539, 0.02958116, 0.03017876, 0.03078843, 0.03141042,
       0.03204497, 0.03269235, 0.0333528 , 0.03402659, 0.034714  ,
       0.03541529, 0.03613075, 0.03686067, 0.03836503, 0.03914008,
       0.03993079, 0.04073747, 0.04156045, 0.04240006, 0.04325662,
       0.04413049, 0.04502202, 0.04593155, 0.04685946, 0.04780612,
       0.0487719 , 0.04975719, 0.05076239, 0.02483832])


def test_simulate_tso_photometry():
    # With NIRSpec / imag
    pando = jwst.PandeiaCalculation('nircam', 'lw_ts')
    scene = jwst.make_scene('phoenix', 'k5v', '2mass,ks', 13.5)
    pando.calc['scene'] = [scene]

    filters = pando.get_configs('filters')
    filter = 'f470n'
    subarray = 'sub160p'
    ngroup = 6

    transit_dur = 2.71
    obs_dur = 6.0
    obs_type = 'transit'
    photo = []
    for filter in filters:
        tso = pando.tso_calculation(
            obs_type, transit_dur, obs_dur, depth_model,
            ngroup, filter=filter, subarray=subarray,
        )
        sim = jwst.simulate_tso(tso, n_obs=1, noiseless=True)
        photo.append(sim)


    # With MIRI / imaging
    import gen_tso.pandeia_io as jwst

    pando = jwst.PandeiaCalculation('miri', 'imaging_ts')
    scene = jwst.make_scene('phoenix', 'k5v', '2mass,ks', 13.5)
    pando.calc['scene'] = [scene]

    filters = pando.get_configs('filters')
    filter = 'f560w'
    ngroup = 6

    model_path = f'{ROOT}data/models//WASP80b_transit.dat'
    depth_model = np.loadtxt(model_path, unpack=True)
    transit_dur = 2.71
    obs_dur = 6.0
    obs_type = 'transit'
    photo = []
    for filter in filters:
        tso = pando.tso_calculation(
            obs_type, transit_dur, obs_dur, depth_model,
            ngroup, filter=filter,
        )
        sim = jwst.simulate_tso(tso, n_obs=1, noiseless=True)
        photo.append(sim)


    plt.figure(10)
    plt.clf()
    plt.plot(depth_model[0], depth_model[1], c='0.6')
    plt.xlim(2.0, 5.8)
    for i,sim in enumerate(photo):
        bin_wl, bin_spec, bin_err, bin_widths = sim
        col = plt.cm.rainbow(i/len(filters))
        plt.errorbar(
            bin_wl, bin_spec, yerr=bin_err, xerr=bin_widths/2,
            fmt='o', color=col, mec='k', mew=1.0, label=filters[i],
        )
    plt.legend(loc='best')


def get_band_width(inst, mode, aperture, filter):
    """
    import gen_tso.pandeia_io as jwst

    inst = 'miri'
    mode = 'imaging_ts'
    pando = jwst.PandeiaCalculation(inst, mode)
    aperture = pando.calc['configuration']['instrument']['aperture']
    filters = pando.get_configs('filters')

    inst = 'nircam'
    mode = 'lw_ts'
    pando = jwst.PandeiaCalculation(inst, mode)
    aperture = pando.calc['configuration']['instrument']['aperture']
    filters = pando.get_configs('filters')
    throughputs = jwst.filter_throughputs(inst=inst, mode=mode)
    """

    plt.figure(10)
    plt.clf()
    nfilters = len(filters)
    for i in range(nfilters):
        filter = filters[i]
        wl0, bw, min_wl, max_wl = jwst.get_bandwidths(inst, mode, aperture, filter)
        wl_width = np.array([[wl0 - min_wl, max_wl - wl0]]).T
        passband = throughputs[aperture][filter]
        band_wl = passband['wl']
        response = passband['response']

        #plt.figure(10)
        #plt.clf()
        h = 0.5*np.amax(response)
        col = plt.cm.rainbow(i/nfilters)
        plt.plot(band_wl, response, c=col)
        plt.errorbar(wl0, h, xerr=wl_width, fmt='o', c=col)
        plt.errorbar(wl0, 0.95*h, xerr=0.5*bw, fmt='o', c='k')







# Copyright (c) 2025 Patricio Cubillos
# Gen TSO is open-source software under the GPL-2.0 license (see LICENSE)

"""
A collection of low-level routines to handle catalogs.
"""

__all__ = [
    'esasky_js_circle',
    'esasky_js_catalog',
    'normalize_name',
    'is_letter',
    'is_candidate',
    'get_letter',
    'get_host',
    'select_alias',
    'invert_aliases',
    'to_float',
    'as_str',
]

import re

import numpy as np


def esasky_js_circle(ra, dec, radius, color='#15B01A'):
    """
    Construct a JS command to draw a circle footprint for ESASky

    Parameters
    ----------
    ra: Float
        Right ascention of the center of the circle footprint (deg).
    dec: Float
        Declination of the center of the circle footprint (deg).
    radius: Float
        Radius of the circle footprint (deg).

    Returns
    -------
    footprint: Dictionary
        A dictionary with the command to draw a circle footprint
        when converted to JSON format, e.g.:
        command = json.dumps(footprint)

    For details on the ESASky JS API see:
    https://www.cosmos.esa.int/web/esdc/esasky-javascript-api

    Examples
    --------
    >>> import gen_tso.catalogs.utils as u
    >>> ra = 315.0259661
    >>> dec = -5.094857
    >>> radius = 80.0
    >>> circle = u.esasky_js_circle(ra, dec, radius)
    """
    footprint = {
        'event': 'overlayFootprints',
        'content': {
            'overlaySet': {
                'type': 'FootprintListOverlay',
                'overlayName': 'visit splitting distance',
                'cooframe': 'J2000',
                'color': color,
                'lineWidth': 5,
                'skyObjectList': [
                    {'name': 'visit splitting distance',
                     'id': 1,
                     'stcs': f'CIRCLE ICRS {ra:.8f} {dec:.8f} {radius/3600:.4f}',
                     'ra_deg': f'{ra:.8f}',
                     'dec_deg': f'{dec:.8f}',
                    }
                ]
            }
        }
    }
    return footprint


def json_target_property(name, value, format):
    """
    Create a json dictionary of a target's property (to be used on
    an overlayCatalog for ESASky).
    """
    prop = {
        'name': name,
        'value': f'{value:{format}}',
        'type': 'STRING'
    }
    return prop


def json_target(index, name, ra, dec, g_mag, teff, logg, separation):
    """
    Create a json dictionary of a target (to be used on an overlayCatalog
    for ESASky).
    """
    data = [
        json_target_property('G mag', g_mag[index], '.2f'),
        json_target_property('T eff', teff[index], '.1f'),
        json_target_property('log(g)', logg[index], '.2f'),
        json_target_property('Separation', separation[index], '.3f'),
    ]

    target = {
        'name': name[index],
        'id': index+1,
        'ra': f'{ra[index]:.8f}',
        'dec': f'{dec[index]:.8f}',
        'data': data,
    }
    return target


def esasky_js_catalog(query):
    """
    Construct a JS command to draw an overlayCatalog footprint for ESASky

    Parameters
    ----------
    query: List of arrays
        A list of arrays containing the names, g_mag, teff, logg,
        ra, dec, and separation of a set of targets (see Examples).

    Returns
    -------
    command: Dictionary
        A dictionary with the command to draw an overlayCatalog
        when converted to JSON format, e.g.:
        js_command = json.dumps(command)

    For details on the ESASky JS API see:
    https://www.cosmos.esa.int/web/esdc/esasky-javascript-api

    Examples
    --------
    >>> import gen_tso.catalogs as cat
    >>> import gen_tso.catalogs.utils as u

    >>> # Stellar sources around WASP-69:
    >>> ra_source = 315.0259661
    >>> dec_source = -5.094857
    >>> query = cat.fetch_gaia_targets(ra_source, dec_source)
    >>> circle = u.esasky_js_catalog(query)
    """
    names, g_mag, teff, logg, ra, dec, separation = query
    ntargets = len(names)
    targets = []
    for i in range(ntargets):
        target = json_target(i, names, ra, dec, g_mag, teff, logg, separation)
        targets.append(target)

    command = {
        "event": 'overlayCatalogue',
        'content': {
            'overlaySet': {
                'type': 'SourceListOverlay',
                'overlayName': 'Nearby Gaia sources',
                'cooframe': 'J2000',
                'color': '#ee2345',
                'lineWidth': 10,
                'skyObjectList': targets,
            }
        }
    }
    return command


def normalize_name(target):
    """
    Normalize target names into a 'more standard' format.
    Mainly to resolve trexolists target names.
    """
    name = re.sub(r'\s+', ' ', target)
    # It's a case issue:
    name = name.replace('KEPLER', 'Kepler')
    name = name.replace('TRES', 'TrES')
    name = name.replace('WOLF-', 'Wolf ')
    name = name.replace('HATP', 'HAT-P-')
    name = name.replace('AU-MIC', 'AU Mic')

    # Custom correction before going over prefixes
    if name.startswith('NAME-'):
        name = name[5:]
    # Prefixes
    name = name.replace('GL', 'GJ')
    prefixes = [
        'L', 'G', 'HD', 'GJ', 'LTT', 'LHS', 'HIP', 'WD',
        'LP', '2MASS', 'PSR', 'IRAS', 'TYC', 'TIC', 'PSO',
    ]
    for prefix in prefixes:
        prefix_len = len(prefix)
        if name.startswith(prefix) and not name[prefix_len].isalpha():
            name = name.replace(f'{prefix}-', f'{prefix} ')
            name = name.replace(f'{prefix}_', f'{prefix} ')
            if name[prefix_len] != ' ':
                name = f'{prefix} ' + name[prefix_len:]

    prefixes = ['CD-', 'BD-', 'BD+']
    for prefix in prefixes:
        prefix_len = len(prefix)
        dash_loc = name.find('-', prefix_len)
        if name.startswith(prefix) and dash_loc > 0:
            name = name[0:dash_loc] + ' ' + name[dash_loc+1:]
    # Main star
    if name.endswith('A') and not name[-2].isspace():
        name = name[:-1] + ' A'
    # Planet letter is in the name
    if name.lower().endswith('b') and not name[-2].isalpha():
        name = name[:-1]
    if name.lower().endswith('d') and not name[-2].isalpha():
        name = name[:-1]

    # Custom corrections
    name = name.replace('-offset', '')
    name = name.replace('-updated', '')
    name = name.replace('-copy', '')
    name = name.replace('-revised', '')
    if name.endswith('-'):
        name = name[:-1]

    if name.upper() in ['55CNC', 'RHO01-CNC', '-RHO01-CNC']:
        name = '55 Cnc'
    if name == 'WD 1856':
        name = 'WD 1856+534'
    if 'V1298' in name:
        name = 'V1298 Tau'
    return name


def is_letter(name):
    """
    Check if name ends with a blank + lower-case letter (it's a planet)
    """
    return name[-1].islower() and name[-2] == ' '


def is_candidate(name):
    """
    Check if name ends with a blank + lower-case letter (it's a planet)
    """
    return len(name)>=3 and name[-3] == '.' and name[-2:].isnumeric()


def get_letter(name):
    """
    Extract 'letter' identifier for a planet name.
    Valid confirmed planet names end with a lower-case letter preceded
    by a blank.  Valid planet candidate names end with a dot followed
    by two numbers.

    Examples
    --------
    >>> get_letter('TOI-741.01')
    >>> get_letter('WASP-69 b')
    """
    if is_letter(name):
        return name[-2:]
    if '.' in name:
        idx = name.rfind('.')
        return name[idx:]
    return ''


def get_host(name):
    """
    Extract host name from a given planet name.
    Valid confirmed planet names end with a lower-case letter preceded
    by a blank.  Valid planet candidate names end with a dot followed
    by two numbers.

    Examples
    --------
    >>> get_host('TOI-741.01')
    >>> get_host('WASP-69 b')
    """
    if is_letter(name):
        return name[:-2]
    if '.' in name:
        idx = name.rfind('.')
        return name[:idx]
    return ''


def select_alias(aka, catalogs, default_name=None):
    """
    Search alternative names take first one found in catalogs list.
    """
    for catalog in catalogs:
        for alias in aka:
            if alias.startswith(catalog):
                return alias
    return default_name


def invert_aliases(aliases):
    """
    Invert an {alias:name} dictionary into {name:aliases_list}
    """
    aka = {}
    for key,val in aliases.items():
        if val not in aka:
            aka[val] = []
        aka[val].append(key)
    return aka


def to_float(value):
    """
    Cast string to None or float type.
    """
    if value == 'None':
        return None
    return float(value)


def as_str(val, fmt='', if_none=None):
    """
    Format as string
    """
    if val is None or np.isnan(val):
        return if_none
    return f'{val:{fmt}}'




# Copyright (c) 2025 Patricio Cubillos
# Gen TSO is open-source software under the GPL-2.0 license (see LICENSE)

import numpy as np
import pyratbay.spectrum as ps
from gen_tso import pandeia_io as jwst
from gen_tso.pandeia_io.pandeia_defaults import (
    _load_flux_rate_splines,
    get_detector,
    make_saturation_label,
)


detectors = jwst.generate_all_instruments()
throughputs = jwst.get_throughputs()

# Pre-computed flux rates
flux_rate_splines, full_wells = _load_flux_rate_splines()

# Catalog of stellar SEDs:
p_keys, p_models, p_teff, p_logg = jwst.get_sed_list('phoenix')
k_keys, k_models, k_teff, k_logg = jwst.get_sed_list('k93models')

phoenix_dict = {key:model for key,model in zip(p_keys, p_models)}
kurucz_dict = {key:model for key,model in zip(k_keys, k_models)}
sed_dict = {
    'phoenix': phoenix_dict,
    'kurucz': kurucz_dict,
}

bands_dict = {
    '2mass,j': 'J mag',
    '2mass,h': 'H mag',
    '2mass,ks': 'Ks mag',
    'gaia,g': 'Gaia mag',
    'johnson,v': 'V mag',
}


def get_throughput(input, evaluate=False):
    config = parse_instrument(
        input, 'instrument', 'mode',
        'aperture', 'disperser', 'filter', 'subarray', 'detector',
    )
    if config is None:
        return None
    inst, mode, aperture, disperser, filter, subarray, detector = config
    obs_type = detector.obs_type

    key = aperture if obs_type == 'photometry' else subarray
    if key not in throughputs[obs_type][inst][mode]:
        return None

    if mode == 'lrsslitless':
        filter = 'None'
    elif mode == 'mrs_ts':
        filter = disperser
    elif mode == 'bots':
        filter = f'{disperser}/{filter}'

    if evaluate:
        return throughputs[obs_type][inst][mode][key][filter]
    config = inst, mode, key, filter
    return throughputs[obs_type], config


def get_auto_sed(input):
    """
    Guess the model closest to the available options given a T_eff
    and log_g pair.
    """
    sed_type = input.sed_type()
    sed_models = sed_dict[sed_type]
    if sed_type == 'kurucz':
        sed_type = 'k93models'

    try:
        t_eff = float(input.t_eff.get())
        log_g = float(input.log_g.get())
    except ValueError:
        return sed_models, None
    chosen_sed = jwst.find_closest_sed(t_eff, log_g, sed_type)
    return sed_models, chosen_sed


def get_saturation_values(
        inst, mode, aperture, disperser, filter, subarray, order,
        sed_label, norm_mag,
        cache_saturation,
    ):
    """
    Get pixel_rate and full_well from instrumental settings.
    """
    sat_label = make_saturation_label(
        inst, mode, aperture, disperser, filter, subarray, order, sed_label,
    )

    sed_items = sat_label.split('_')
    band_label = sed_items[-1]
    sat_guess_label = '_'.join(sed_items[0:-2])
    can_guess = band_label == 'Ks' and sat_guess_label in flux_rate_splines
    pixel_rate = None
    full_well = None
    if sat_label in cache_saturation:
        pixel_rate = cache_saturation[sat_label]['brightest_pixel_rate']
        full_well = cache_saturation[sat_label]['full_well']
    elif can_guess:
        cs = flux_rate_splines[sat_guess_label]
        pixel_rate = 10**cs(norm_mag)
        full_well = full_wells[sat_guess_label]
    return pixel_rate, full_well


def draw(tso_list, resolution, n_obs):
    """
    Draw a random noised-up transit/eclipse depth realization from a TSO
    """
    if not isinstance(tso_list, list):
        tso_list = [tso_list]

    sims = []
    for tso in tso_list:
        bin_wl, bin_spec, bin_err, wl_widths = jwst.simulate_tso(
           tso, n_obs=n_obs, resolution=resolution, noiseless=False,
        )
        sims.append({
            'wl': bin_wl,
            'depth': bin_spec,
            'uncert': bin_err,
            'wl_widths': wl_widths,
        })
    return sims


def planet_model_name(input):
    """
    Get the planet model name based on the transit/eclipse depth values.

    Returns
    -------
    depth_label: String
        A string representation of the depth model.
    """
    planet_model_type = input.planet_model_type.get()
    if planet_model_type == 'Input':
        return input.depth.get()
    elif planet_model_type == 'Flat':
        transit_depth = input.transit_depth.get()
        return f'Flat transit ({transit_depth:.3f}%)'
    elif planet_model_type == 'Blackbody':
        eclipse_depth = input.eclipse_depth.get()
        t_planet = input.teq_planet.get()
        return f'Blackbody({t_planet:.0f}K, rprs\u00b2={eclipse_depth:.3f}%)'


def parse_instrument(input, *args):
    """
    Parse instrumental configuration from front-end to back-end.
    Ensure that only the requested parameters are a valid configuration.
    """
    # instrument and mode always checked
    inst = input.instrument.get().lower()
    mode = input.mode.get()
    detector = get_detector(inst, mode, detectors)
    if detector is None:
        return None

    config = {
        'instrument': inst,
        'mode': mode,
        'detector': detector,
    }

    if 'aperture' in args:
        aperture = input.aperture.get()
        has_pupils = mode in ['lw_ts', 'sw_ts']
        if has_pupils and aperture not in detector.pupils:
            return None
        if not has_pupils and aperture not in detector.apertures:
            return None
        if has_pupils:
            aperture = detector.pupil_to_aperture[aperture]
        config['aperture'] = aperture

    if 'disperser' in args:
        disperser = input.disperser.get()
        if disperser not in detector.dispersers:
            return None
        config['disperser'] = disperser

    if 'filter' in args:
        filter = input.filter.get()
        if filter not in detector.filters:
            return None
        config['filter'] = filter

    if 'subarray' in args:
        subarray = input.subarray.get()
        if subarray not in detector.subarrays:
            return None
        config['subarray'] = subarray

    if 'readout' in args:
        readout = input.readout.get()
        if readout not in detector.readouts:
            return None
        config['readout'] = readout

    # Now parse front-end to back-end:
    if 'pairing' in args:
        if mode == 'sw_ts':
            config['pairing'] = input.pairing.get()
        else:
            config['pairing'] = None

    if 'pupil' in args:
        config['pupil'] = input.aperture.get()

    if 'ngroup' in args:
        if mode == 'target_acq':
            ngroup = int(input.ngroup_acq.get())
            config['disperser'] = None
        else:
            ngroup = input.ngroup.get()
        config['ngroup'] = ngroup

    config['nint'] = 1 if mode == 'target_acq' else input.integrations.get()

    if mode == 'mrs_ts':
        config['aperture'] = ['ch1', 'ch2', 'ch3', 'ch4']

    if mode == 'bots' and ('disperser' in args or 'filter' in args):
        if 'filter' not in args:
            filter = input.filter.get()
        config['disperser'], config['filter'] = filter.split('/')

    if 'order' in args:
        if mode == 'soss':
            if filter == 'f277w':
                order = [1]
            else:
                order = input.order.get()
                order = [int(val) for val in order.split()]
        else:
            order = None
        config['order'] = order

    # Return in the same order as requested
    config_list = [config[arg] for arg in args]

    return config_list


def parse_depth_model(input, spectra):
    """
    Parse transit/eclipse model name based on current state.
    Calculate or extract model.
    """
    model_type = input.planet_model_type.get()
    depth_label = planet_model_name(input)
    obs_geometry = input.obs_geometry.get()

    if model_type == 'Input':
        if depth_label is None:
            wl, depth = None, None
        else:
            wl = spectra[obs_geometry][depth_label]['wl']
            depth = spectra[obs_geometry][depth_label]['depth']
    elif model_type == 'Flat':
        transit_depth = input.transit_depth.get() * 0.01
        wl = ps.constant_resolution_spectrum(0.1, 50.0, resolution=300)
        nwave = len(wl)
        depth = np.tile(transit_depth, nwave)
    elif model_type == 'Blackbody':
        rprs = np.sqrt(input.eclipse_depth.get() * 0.01)
        t_planet = input.teq_planet.get()
        sed_type, sed = parse_sed(input, spectra)[0:2]
        wl, depth = jwst.blackbody_eclipse_depth(t_planet, rprs, sed_type, sed)

    return depth_label, wl, depth


def parse_obs(input):
    planet_model_type = input.planet_model_type.get()
    depth_model = None
    rprs_sq = None
    teq_planet = None
    if planet_model_type == 'Input':
        depth_model = input.depth.get()
    elif planet_model_type == 'Flat':
        rprs_sq = input.transit_depth.get()
    elif planet_model_type == 'Blackbody':
        rprs_sq = input.eclipse_depth.get()
        teq_planet = input.teq_planet.get()
    return planet_model_type, depth_model, rprs_sq, teq_planet


def parse_sed(input, spectra, target_acq_mag=None):
    """Extract SED parameters"""
    if target_acq_mag is None:
        sed_type = input.sed_type()
        norm_band = input.magnitude_band.get()
        norm_magnitude = float(input.magnitude.get())
    else:
        sed_type = 'phoenix'
        norm_band = 'gaia,g'
        norm_magnitude = target_acq_mag

    if sed_type in ['phoenix', 'kurucz']:
        if target_acq_mag is None:
            sed_model = input.sed.get()
        else:
            sed_model = input.ta_sed.get()
        if sed_model not in sed_dict[sed_type]:
            return None, None, None, None, None
        model_label = f'{sed_type}_{sed_model}'
    elif sed_type == 'blackbody':
        sed_model = float(input.t_eff.get())
        model_label = f'bb_{sed_model:.0f}K'
    elif sed_type == 'input':
        model_label = input.sed.get()
        if model_label not in spectra['sed']:
            return None, None, None, None, None
        sed_model = spectra['sed'][model_label]

    if sed_type == 'kurucz':
        sed_type = 'k93models'

    # Make a label
    band_name = bands_dict[norm_band].split()[0]
    band_label = f'{norm_magnitude:.2f}_{band_name}'
    sed_label = f'{model_label}_{band_label}'

    return sed_type, sed_model, norm_band, norm_magnitude, sed_label




# Copyright (c) 2025 Patricio Cubillos
# Gen TSO is open-source software under the GPL-2.0 license (see LICENSE)

__all__ = [
    'check_latest_pandeia_version',
    'check_pandeia_ref_data',
    'check_pysynphot',
    'update_synphot_files',
    'fetch_synphot_files',
]

import os
from pathlib import Path
import shutil
import tarfile

import prompt_toolkit as ptk
import requests
from shiny import ui


stsci_url = 'https://archive.stsci.edu/hlsps/reference-atlases/'
syn1 = 'hlsp_reference-atlases_hst_multi_everything_multi_v16_sed.tar'
syn4 = 'hlsp_reference-atlases_hst_multi_kurucz-1993-atlas_multi_v2_synphot4.tar'
syn5 = 'hlsp_reference-atlases_hst_multi_pheonix-models_multi_v3_synphot5.tar'


def check_latest_pandeia_version():
    response = requests.get('https://pypi.org/pypi/pandeia.engine/json')
    last_pandeia = response.json()['info']['version']
    return last_pandeia


def check_pandeia_ref_data(latest_version):
    """
    Check that the Pandeia reference data environment variable exists.
    Check that the path exists.
    Check that the reference data version is consistent with Pandeia.engine

    Returns
    -------
    ui.HTML() formatted text with state of the pandeia reference data.
    """
    pandeia_url = (
        "https://outerspace.stsci.edu/display/PEN/Pandeia+Engine+Installation"
    )

    if "pandeia_refdata" not in os.environ:
        output = "Missing '$pandeia_refdata' environment variable"
    else:
        refdata_path = os.environ['pandeia_refdata']
        try:
            with open(f"{refdata_path}/VERSION_PSF") as fp:
                data_version = fp.readline().strip()
            if data_version == latest_version:
                output = (
                    f"Pandeia reference data version {data_version} is up to "
                    f"date<br>$pandeia_refdata={refdata_path}"
                )
            else:
                output = (
                    f"Pandeia reference data version: {data_version} < "
                    f"pandeia.engine version {latest_version}"
                )
        except OSError:
            output = f"Invalid 'pandeia_refdata' path: {repr(refdata_path)}"

    if 'up to date' in output:
        return ui.HTML(f'<span style="color:#0B980D">{output}</span>')

    return ui.span(
        ui.HTML(
            f'<span style="color:red">{output}.</span> '
            'Please follow the instructions in section 3.1 of '
        ),
        ui.tags.a(pandeia_url, href=pandeia_url, target="_blank"),
    )


def check_pysynphot():
    """
    Check pysynphot environment variable exists.
    Check that the path exists.
    Check that the required files are there.

    Returns
    -------
    ui.HTML() formatted text with state of pysynphot.
    """
    if "PYSYN_CDBS" not in os.environ:
        return "Missing 'PYSYN_CDBS' path"

    cdbs_path = os.environ['PYSYN_CDBS']
    if not os.path.exists(cdbs_path):
        return f'"PYSYN_CDBS" path does not exist:<br>{cdbs_path}'

    # check files
    atlases = {
        'throughput': os.path.exists(f'{cdbs_path}/comp'),
        'Kurucz SED': os.path.exists(f'{cdbs_path}/grid/k93models/'),
        'PHOENIX SED': os.path.exists(f'{cdbs_path}/grid/phoenix/'),
    }

    lost = []
    found = []
    for atlas, is_there in atlases.items():
        if is_there:
            found.append(atlas)
        else:
            lost.append(atlas)
    found_txt = ', '.join(found)
    lost_txt = ', '.join(lost)

    if len(lost) == 0:
        return ui.HTML(
            f'<span style="color:#0B980D">All {found_txt} files are '
            'in place</span>'
        )
    if len(found) == 0:
        return ui.HTML(
            f'<span style="color:red">Could not find {lost_txt} files</span>'
        )

    return ui.HTML(
        f'<p><span style="color:#0B980D">Found {found_txt} files.</span> '
        f'<span style="color:red">But {lost_txt} files are missing.</span></p>'
    )


def update_synphot_files(force_update=False):
    """
    Check which synphot files are needed, fetch them from STScI,
    https://archive.stsci.edu/hlsp/reference-atlases
    and save them to the right folder pointed by the $PYSYN_CDBS
    environment variable.

    Parameters
    ----------
    force_update: Bool
        If True, enforce updating files even if they already exist.

    Returns
    -------
    warnings: List of strings
        List of any error during the file requests (if any).

    Examples
    --------
    >>> from gen_tso.pandeia_io.pandeia_setup import fetch_synphot_files
    >>> fetch_synphot_files('k93models')
    """
    unset_pysyn = (
        "'$PYSYN_CDBS' environment variable was not set. "
        "Could not setup pysynphot reference data"
    )
    if "PYSYN_CDBS" not in os.environ:
        completer = ptk.completion.PathCompleter(
            only_directories=True,
            expanduser=True,
        )
        synphot_dir = ptk.prompt(
            "The '$PYSYN_CDBS' environment variable is not defined.\nPlease "
            "set it to a dir where to store the Pandeia synphot data "
            "(~3GB of data)\n(and make sure the top folder is named 'trds')\n"
            "$PYSYN_CDBS = ",
            completer=completer,
        )

        if synphot_dir.strip() == '':
            print(unset_pysyn)
            return

        path = Path(os.path.realpath(os.path.expanduser(synphot_dir)))
        if path.name != 'trds':
            go_on = ptk.prompt(
                "The '$PYSYN_CDBS' top dir is not named 'trds', what about:"
                f"\n{path / 'trds'}"
                '\nProceed (Y/n)? '
            )
            if go_on.strip().lower() not in ['', 'y', 'yes']:
                print(unset_pysyn)
                return
            path = path / 'trds'
        print(
            "\nNow, add this line to your bash file:\n"
            f"export PYSYN_CDBS={path}\nand then 'source' "
            "the bash file or launch the app from a new terminal"
        )
        synphot_path = str(path)
    else:
        synphot_path = os.environ['PYSYN_CDBS']

    if not os.path.exists(synphot_path):
        Path(synphot_path).mkdir(parents=True, exist_ok=True)
        if not os.path.exists(synphot_path):
            raise ValueError(
                f'Could not create "PYSYN_CDBS" directory: {synphot_path}'
            )

    # Check files to update
    atlases = {
        'throughput': os.path.exists(f'{synphot_path}/comp'),
        'Kurucz SED': os.path.exists(f'{synphot_path}/grid/k93models/'),
        'PHOENIX SED': os.path.exists(f'{synphot_path}/grid/phoenix/'),
    }

    warnings = []
    if not atlases['Kurucz SED'] or force_update:
        warnings.append(fetch_synphot_files('k93models', synphot_path))
    if not atlases['PHOENIX SED'] or force_update:
        warnings.append(fetch_synphot_files('phoenix', synphot_path))
    if not atlases['throughput'] or force_update:
        warnings.append(fetch_synphot_files('throughput', synphot_path))

    warnings = [warn for warn in warnings if warn is not None]
    return warnings


def fetch_synphot_files(synphot_file, synphot_path=None):
    """
    Download Kurucz or PHOENIX SED tar files from STScI.
    Unzip tar file.
    Place k93models/ or phoenix/ folder into $PYSYN_CDBS/trds/grid/

    Parameters
    ----------
    sed: String
        Select from 'k93models', 'phoenix', or 'throughput'.

    Returns
    -------
    On success, returns None.
    If there was an error, return a string describing the error.

    Examples
    --------
    >>> from gen_tso.pandeia_io.pandeia_setup import fetch_synphot_files
    >>> fetch_synphot_files('k93models')
    """
    if synphot_path is None:
        if "PYSYN_CDBS" not in os.environ:
            return '$PYSYN_CDBS environment variable is not defined'
        synphot_path = os.environ['PYSYN_CDBS']

    if synphot_file == 'k93models':
        url = f'{stsci_url}{syn4}'
    elif synphot_file == 'phoenix':
        url = f'{stsci_url}{syn5}'
    elif synphot_file == 'throughput':
        url = f'{stsci_url}{syn1}'
    else:
        return f'Invalid Synphot file type: {synphot_file}'

    if not os.path.exists(synphot_path):
        Path(synphot_path).mkdir(parents=True, exist_ok=True)

    print(f'Downloading {synphot_file} files ...')
    tmp_tar_file = f'{synphot_path}/synphot_file_{synphot_file}.tar'
    with requests.get(url, stream=True) as response:
        if not response.ok:
            return (
                'HTTP request failed. '
                f'Could not download {synphot_file} reference files'
            )
        with open(tmp_tar_file, mode="wb") as file:
            for chunk in response.iter_content(chunk_size=10 * 1024):
                file.write(chunk)

    print('Saving to PYSYN_CDBS grid folder')
    tmp_dir = f'{synphot_path}/tmp_synphot'
    if os.path.exists(tmp_dir):
        shutil.rmtree(tmp_dir)
    with tarfile.open(tmp_tar_file, "r") as tar:
        tar.extractall(path=tmp_dir)
    shutil.copytree(
        f'{tmp_dir}/grp/redcat/trds/', synphot_path,
        dirs_exist_ok=True,
    )
    shutil.rmtree(tmp_dir)
    os.remove(tmp_tar_file)
    print(f'{synphot_file} files updated!')




# Copyright (c) 2025 Patricio Cubillos
# Gen TSO is open-source software under the GPL-2.0 license (see LICENSE)

__all__ = [
    'filter_popover',
    'sed_popover',
    'planet_popover',
    'tso_popover',
]

import faicons as fa
from shiny import ui


depth_units = [
    "none",
    "percent",
    "ppm",
]

wl_scales = {
    'Wavelength scale': {
        'linear': 'linear',
        'log': 'log',
    },
}

tso_choices = {
    'tso': 'TSO',
    'fluxes': 'Flux rate',
    'snr': 'S/N',
    '2d_flux': '2D flux',
    '2d_snr': '2D S/N',
    '2d_saturation': '2D saturation',
    '2d_groups': '2D groups',
}


filter_popover = ui.popover(
    ui.span(
        fa.icon_svg("gear"),
        style="position:absolute; top: 5px; right: 7px;",
    ),
    "Show filter throughputs",
    ui.input_radio_buttons(
        id="filter_filter",
        label=None,
        choices=["none", "all"],
        inline=True,
        selected="none",
    ),
    placement="top",
    id="filter_popover",
)


sed_popover = ui.popover(
    ui.span(
        fa.icon_svg("gear"),
        style="position:absolute; top: 5px; right: 7px;",
    ),
    ui.layout_column_wrap(
        ui.input_numeric(
            id='plot_sed_resolution',
            label='Resolution:',
            value=0.0,
            min=10.0, max=3000.0, step=25.0,
        ),
        ui.input_select(
            id="plot_sed_units",
            label="Flux units:",
            choices=['mJy'],
            selected='mJy',
        ),
        width=1/2,
        fixed_width=False,
        gap='5px',
        fill=False,
        fillable=True,
        class_="p-0 pb-1 m-0",
    ),
    ui.layout_column_wrap(
        "Wavelength:",
        ui.input_numeric(
            id='sed_wl_min', label='',
            value=0.45, min=0.3, max=30.0, step=0.15,
        ),
        ui.input_numeric(
            id='sed_wl_max', label='',
            value=28.0, min=0.5, max=30.0, step=1.0,
        ),
        ui.input_select(
            id="plot_sed_xscale",
            label="",
            choices=wl_scales,
            selected='log',
        ),
        width=1/4,
        fixed_width=False,
        gap='5px',
        fill=False,
        fillable=True,
        class_="p-0 m-0",
    ),
    placement="top",
    id="sed_popover",
)


planet_popover = ui.popover(
    ui.span(
        fa.icon_svg("gear"),
        style="position:absolute; top: 5px; right: 7px;",
    ),
    ui.layout_column_wrap(
        ui.input_numeric(
            id='depth_resolution',
            label='Resolution:',
            value=250.0,
            min=10.0, max=3000.0, step=25.0,
        ),
        ui.input_select(
            id="plot_depth_units",
            label="Depth units:",
            choices=depth_units,
            selected='percent',
        ),
        width=1/2,
        fixed_width=False,
        gap='5px',
        fill=False,
        fillable=True,
        class_="p-0 pb-1 m-0",
    ),
    ui.layout_column_wrap(
        "Wavelength:",
        ui.input_numeric(
            id='depth_wl_min', label='',
            value=0.6, min=0.3, max=30.0, step=0.15,
        ),
        ui.input_numeric(
            id='depth_wl_max', label='',
            value=28.0, min=0.5, max=30.0, step=1.0,
        ),
        ui.input_select(
            "plot_depth_xscale",
            label="",
            choices=wl_scales,
            selected='log',
        ),
        width=1/4,
        fixed_width=False,
        gap='5px',
        fill=False,
        fillable=True,
        class_="p-0 m-0",
    ),
    placement="top",
    id="depth_popover",
)


tso_popover = ui.popover(
    ui.span(
        fa.icon_svg("gear"),
        style="position:absolute; top: 5px; right: 7px;",
    ),
    ui.layout_column_wrap(
        ui.input_select(
            id="tso_plot",
            label="Plot:",
            choices=tso_choices,
            selected='tso',
        ),
        ui.panel_conditional(
            "input.tso_plot == 'tso'",
            ui.input_numeric(
                id='n_obs',
                label='Number of obs:',
                value=1.0,
                min=1.0, max=3000.0, step=1.0,
                width='200px',
            ),
        ),
        ui.panel_conditional(
            "input.tso_plot == 'tso'",
            ui.input_numeric(
                id='tso_resolution',
                label='Resolution:',
                value=250.0,
                min=25.0, max=3000.0, step=25.0,
                width='200px',
            ),
        ),
        ui.panel_conditional(
            "input.tso_plot == 'tso'",
            ui.input_select(
                id="plot_tso_units",
                label="Depth units:",
                choices = depth_units,
                selected='percent',
            ),
        ),
        width=1/4,
        fixed_width=False,
        gap='5px',
        fill=False,
        fillable=True,
        class_="px-0 py-1 m-0",
    ),
    ui.layout_column_wrap(
        "Wavelength:",
        ui.input_numeric(
            id='tso_wl_min', label='',
            value=None, min=0.5, max=30.0, step=0.1,
        ),
        ui.input_numeric(
            id='tso_wl_max', label='',
            value=None, min=0.5, max=30.0, step=0.1,
        ),
        ui.input_select(
            id="plot_tso_xscale",
            label='',
            choices=wl_scales,
            selected='linear',
        ),
        width=1/4,
        fixed_width=False,
        gap='5px',
        fill=False,
        fillable=True,
        class_="p-0 m-0",
    ),
    ui.panel_conditional(
        "input.tso_plot == 'tso'",
        ui.layout_column_wrap(
            "Depth:",
            ui.input_numeric(
                id='tso_depth_min',
                label='',
                value=None,
            ),
            ui.input_numeric(
                id='tso_depth_max',
                label='',
                value=None,
            ),
            ui.input_action_button(
                id="redraw_tso",
                label="Re-draw",
                class_="btn btn-outline-primary btn-sm",
            ),
            width=1/4,
            fixed_width=False,
            gap='5px',
            fill=False,
            fillable=True,
            class_="p-0 m-0",
        ),
    ),
    placement="top",
    id="tso_popover",
)
